@@ -1,4 +1,4 @@
-# Copyright 2009-2015 MongoDB, Inc.
+# Copyright 2009-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -25,15 +25,24 @@ import random
 import struct
 
 import bson
+from bson import CodecOptions
 from bson.codec_options import DEFAULT_CODEC_OPTIONS
-from bson.py3compat import b, StringIO, u
+from bson.py3compat import b, StringIO
 from bson.son import SON
+
 try:
     from pymongo import _cmessage
     _use_c = True
 except ImportError:
     _use_c = False
-from pymongo.errors import DocumentTooLarge, InvalidOperation, OperationFailure
+from pymongo.errors import (ConfigurationError,
+                            CursorNotFound,
+                            DocumentTooLarge,
+                            ExecutionTimeout,
+                            InvalidOperation,
+                            NotMasterError,
+                            OperationFailure,
+                            ProtocolError)
 from pymongo.read_concern import DEFAULT_READ_CONCERN
 from pymongo.read_preferences import ReadPreference
 
@@ -61,7 +70,10 @@ _OP_MAP = {
     _DELETE: b'\x04deletes\x00\x00\x00\x00\x00',
 }
 
-_UJOIN = u("%s.%s")
+_UJOIN = u"%s.%s"
+
+_UNICODE_REPLACE_CODEC_OPTIONS = CodecOptions(
+    unicode_decode_error_handler='replace')
 
 
 def _randint():
@@ -73,13 +85,16 @@ def _maybe_add_read_preference(spec, read_preference):
     """Add $readPreference to spec when appropriate."""
     mode = read_preference.mode
     tag_sets = read_preference.tag_sets
+    max_staleness = read_preference.max_staleness
     # Only add $readPreference if it's something other than primary to avoid
     # problems with mongos versions that don't support read preferences. Also,
     # for maximum backwards compatibility, don't add $readPreference for
-    # secondaryPreferred unless tags are in use (setting the slaveOkay bit
-    # has the same effect).
+    # secondaryPreferred unless tags or maxStalenessSeconds are in use (setting
+    # the slaveOkay bit has the same effect).
     if mode and (
-        mode != ReadPreference.SECONDARY_PREFERRED.mode or tag_sets != [{}]):
+        mode != ReadPreference.SECONDARY_PREFERRED.mode
+        or tag_sets != [{}]
+        or max_staleness != -1):
 
         if "$query" not in spec:
             spec = SON([("$query", spec)])
@@ -155,19 +170,8 @@ _MODIFIERS = SON([
     ('$snapshot', 'snapshot')])
 
 
-def _gen_explain_command(
-        coll, spec, projection, skip, limit, batch_size,
-        options, read_concern):
-    """Generate an explain command document."""
-    cmd = _gen_find_command(
-        coll, spec, projection, skip, limit, batch_size, options)
-    if read_concern.level:
-        return SON([('explain', cmd), ('readConcern', read_concern.document)])
-    return SON([('explain', cmd)])
-
-
-def _gen_find_command(coll, spec, projection, skip, limit, batch_size,
-                      options, read_concern=DEFAULT_READ_CONCERN):
+def _gen_find_command(coll, spec, projection, skip, limit, batch_size, options,
+                      read_concern, collation=None):
     """Generate a find command document."""
     cmd = SON([('find', coll)])
     if '$query' in spec:
@@ -192,7 +196,8 @@ def _gen_find_command(coll, spec, projection, skip, limit, batch_size,
         cmd['batchSize'] = batch_size
     if read_concern.level:
         cmd['readConcern'] = read_concern.document
-
+    if collation:
+        cmd['collation'] = collation
     if options:
         cmd.update([(opt, True)
                     for opt, val in _OPTIONS.items()
@@ -216,11 +221,12 @@ class _Query(object):
 
     __slots__ = ('flags', 'db', 'coll', 'ntoskip', 'spec',
                  'fields', 'codec_options', 'read_preference', 'limit',
-                 'batch_size', 'name', 'read_concern')
+                 'batch_size', 'name', 'read_concern', 'collation',
+                 'session', 'client')
 
     def __init__(self, flags, db, coll, ntoskip, spec, fields,
                  codec_options, read_preference, limit,
-                 batch_size, read_concern):
+                 batch_size, read_concern, collation, session, client):
         self.flags = flags
         self.db = db
         self.coll = coll
@@ -232,24 +238,58 @@ class _Query(object):
         self.read_concern = read_concern
         self.limit = limit
         self.batch_size = batch_size
+        self.collation = collation
+        self.session = session
+        self.client = client
         self.name = 'find'
 
-    def as_command(self):
+    def use_command(self, sock_info, exhaust):
+        use_find_cmd = False
+        if sock_info.max_wire_version >= 4:
+            if not exhaust:
+                use_find_cmd = True
+        elif not self.read_concern.ok_for_legacy:
+            raise ConfigurationError(
+                'read concern level of %s is not valid '
+                'with a max wire version of %d.'
+                % (self.read_concern.level,
+                   sock_info.max_wire_version))
+
+        if sock_info.max_wire_version < 5 and self.collation is not None:
+            raise ConfigurationError(
+                'Specifying a collation is unsupported with a max wire '
+                'version of %d.' % (sock_info.max_wire_version,))
+
+        sock_info.validate_session(self.client, self.session)
+
+        return use_find_cmd
+
+    def as_command(self, sock_info):
         """Return a find command document for this query.
 
         Should be called *after* get_message.
         """
-        if '$explain' in self.spec:
+        explain = '$explain' in self.spec
+        cmd = _gen_find_command(
+            self.coll, self.spec, self.fields, self.ntoskip,
+            self.limit, self.batch_size, self.flags, self.read_concern,
+            self.collation)
+        if explain:
             self.name = 'explain'
-            return _gen_explain_command(
-                self.coll, self.spec, self.fields, self.ntoskip,
-                self.limit, self.batch_size, self.flags,
-                self.read_concern), self.db
-        return _gen_find_command(
-            self.coll, self.spec, self.fields, self.ntoskip, self.limit,
-            self.batch_size, self.flags, self.read_concern), self.db
+            cmd = SON([('explain', cmd)])
+        session = self.session
+        if session:
+            cmd['lsid'] = session._use_lsid()
+            # Explain does not support readConcern.
+            if (not explain and session.options.causal_consistency
+                    and session.operation_time is not None):
+                cmd.setdefault(
+                    'readConcern', {})[
+                    'afterClusterTime'] = session.operation_time
+        sock_info.send_cluster_time(cmd, session, self.client)
+        return cmd, self.db
 
-    def get_message(self, set_slave_ok, is_mongos, use_cmd=False):
+    def get_message(self, set_slave_ok, sock_info, use_cmd=False):
         """Get a query message, possibly setting the slaveOk bit."""
         if set_slave_ok:
             # Set the slaveOk bit.
@@ -262,7 +302,7 @@ class _Query(object):
 
         if use_cmd:
             ns = _UJOIN % (self.db, "$cmd")
-            spec = self.as_command()[0]
+            spec = self.as_command(sock_info)[0]
             ntoreturn = -1  # All DB commands return 1 document
         else:
             # OP_QUERY treats ntoreturn of -1 and 1 the same, return
@@ -275,7 +315,7 @@ class _Query(object):
                 else:
                     ntoreturn = self.limit
 
-        if is_mongos:
+        if sock_info.is_mongos:
             spec = _maybe_add_read_preference(spec,
                                               self.read_preference)
 
@@ -287,39 +327,73 @@ class _GetMore(object):
     """A getmore operation."""
 
     __slots__ = ('db', 'coll', 'ntoreturn', 'cursor_id', 'max_await_time_ms',
-                 'codec_options')
+                 'codec_options', 'session', 'client')
 
     name = 'getMore'
 
-    def __init__(self, db, coll, ntoreturn, cursor_id, codec_options,
-                 max_await_time_ms=None):
+    def __init__(self, db, coll, ntoreturn, cursor_id, codec_options, session,
+                 client, max_await_time_ms=None):
         self.db = db
         self.coll = coll
         self.ntoreturn = ntoreturn
         self.cursor_id = cursor_id
         self.codec_options = codec_options
+        self.session = session
+        self.client = client
         self.max_await_time_ms = max_await_time_ms
 
-    def as_command(self):
-        """Return a getMore command document for this query."""
-        return _gen_get_more_command(self.cursor_id, self.coll,
-                                     self.ntoreturn,
-                                     self.max_await_time_ms), self.db
+    def use_command(self, sock_info, exhaust):
+        sock_info.validate_session(self.client, self.session)
+        return sock_info.max_wire_version >= 4 and not exhaust
 
-    def get_message(self, dummy0, dummy1, use_cmd=False):
+    def as_command(self, sock_info):
+        """Return a getMore command document for this query."""
+        cmd = _gen_get_more_command(self.cursor_id, self.coll,
+                                    self.ntoreturn,
+                                    self.max_await_time_ms)
+
+        if self.session:
+            cmd['lsid'] = self.session._use_lsid()
+        sock_info.send_cluster_time(cmd, self.session, self.client)
+        return cmd, self.db
+
+    def get_message(self, dummy0, sock_info, use_cmd=False):
         """Get a getmore message."""
 
         ns = _UJOIN % (self.db, self.coll)
 
         if use_cmd:
             ns = _UJOIN % (self.db, "$cmd")
-            spec = self.as_command()[0]
+            spec = self.as_command(sock_info)[0]
 
             return query(0, ns, 0, -1, spec, None, self.codec_options)
 
         return get_more(ns, self.ntoreturn, self.cursor_id)
 
 
+class _RawBatchQuery(_Query):
+    def use_command(self, socket_info, exhaust):
+        # Compatibility checks.
+        super(_RawBatchQuery, self).use_command(socket_info, exhaust)
+
+        return False
+
+    def get_message(self, set_slave_ok, sock_info, use_cmd=False):
+        # Always pass False for use_cmd.
+        return super(_RawBatchQuery, self).get_message(
+            set_slave_ok, sock_info, False)
+
+
+class _RawBatchGetMore(_GetMore):
+    def use_command(self, socket_info, exhaust):
+        return False
+
+    def get_message(self, set_slave_ok, sock_info, use_cmd=False):
+        # Always pass False for use_cmd.
+        return super(_RawBatchGetMore, self).get_message(
+            set_slave_ok, sock_info, False)
+
+
 class _CursorAddress(tuple):
     """The server address (host, port) of a cursor, with namespace property."""
 
@@ -432,7 +506,19 @@ def query(options, collection_name, num_to_skip,
     data += bson._make_c_string(collection_name)
     data += struct.pack("<i", num_to_skip)
     data += struct.pack("<i", num_to_return)
-    encoded = bson.BSON.encode(query, check_keys, opts)
+    if check_keys:
+        # Temporarily remove $clusterTime to avoid an error from the $-prefix.
+        cluster_time = query.pop('$clusterTime', None)
+        encoded = bson.BSON.encode(query, True, opts)
+        if cluster_time is not None:
+            extra = bson._name_value_to_bson(
+                b"$clusterTime\x00", cluster_time, False, opts)
+            encoded = (
+                bson._PACK_INT(len(encoded) + len(extra))
+                + encoded[4:-1] + extra + b'\x00')
+            query['$clusterTime'] = cluster_time
+    else:
+        encoded = bson.BSON.encode(query, False, opts)
     data += encoded
     max_bson_size = len(encoded)
     if field_selector is not None:
@@ -502,10 +588,11 @@ class _BulkWriteContext(object):
     """A wrapper around SocketInfo for use with write splitting functions."""
 
     __slots__ = ('db_name', 'command', 'sock_info', 'op_id',
-                 'name', 'field', 'publish', 'start_time', 'listeners')
+                 'name', 'field', 'publish', 'start_time', 'listeners',
+                 'session')
 
-    def __init__(
-            self, database_name, command, sock_info, operation_id, listeners):
+    def __init__(self, database_name, command, sock_info, operation_id,
+                 listeners, session):
         self.db_name = database_name
         self.command = command
         self.sock_info = sock_info
@@ -515,6 +602,7 @@ class _BulkWriteContext(object):
         self.name = next(iter(command))
         self.field = _FIELD_MAP[self.name]
         self.start_time = datetime.datetime.now() if self.publish else None
+        self.session = session
 
     @property
     def max_bson_size(self):
@@ -565,6 +653,9 @@ class _BulkWriteContext(object):
     def write_command(self, request_id, msg, docs):
         """A proxy for SocketInfo.write_command that handles event publishing.
         """
+        if self.session:
+            # Update last_use time.
+            self.session._use_lsid()
         if self.publish:
             duration = datetime.datetime.now() - self.start_time
             self._start(request_id, docs)
@@ -697,7 +788,7 @@ if _use_c:
 
 def _do_batched_write_command(namespace, operation, command,
                               docs, check_keys, opts, ctx):
-    """Execute a batch of insert, update, or delete commands.
+    """Create the next batched insert, update, or delete command.
     """
     max_bson_size = ctx.max_bson_size
     max_write_batch_size = ctx.max_write_batch_size
@@ -705,8 +796,6 @@ def _do_batched_write_command(namespace, operation, command,
     # Server guarantees there is enough room: SERVER-10643.
     max_cmd_size = max_bson_size + _COMMAND_OVERHEAD
 
-    ordered = command.get('ordered', True)
-
     buf = StringIO()
     # Save space for message length and request id
     buf.write(_ZERO_64)
@@ -738,60 +827,22 @@ def _do_batched_write_command(namespace, operation, command,
 
     # Where to write list document length
     list_start = buf.tell() - 4
-
     to_send = []
-
-    def send_message():
-        """Finalize and send the current OP_QUERY message.
-        """
-        # Close list and command documents
-        buf.write(_ZERO_16)
-
-        # Write document lengths and request id
-        length = buf.tell()
-        buf.seek(list_start)
-        buf.write(struct.pack('<i', length - list_start - 1))
-        buf.seek(command_start)
-        buf.write(struct.pack('<i', length - command_start))
-        buf.seek(4)
-        request_id = _randint()
-        buf.write(struct.pack('<i', request_id))
-        buf.seek(0)
-        buf.write(struct.pack('<i', length))
-        return ctx.write_command(request_id, buf.getvalue(), to_send)
-
-    # If there are multiple batches we'll
-    # merge results in the caller.
-    results = []
-
     idx = 0
-    idx_offset = 0
-    has_docs = False
     for doc in docs:
-        has_docs = True
         # Encode the current operation
         key = b(str(idx))
         value = bson.BSON.encode(doc, check_keys, opts)
-        # Send a batch?
-        enough_data = (buf.tell() + len(key) + len(value) + 2) >= max_cmd_size
+        # Is there enough room to add this document? max_cmd_size accounts for
+        # the two trailing null bytes.
+        enough_data = (buf.tell() + len(key) + len(value)) >= max_cmd_size
         enough_documents = (idx >= max_write_batch_size)
         if enough_data or enough_documents:
             if not idx:
                 write_op = "insert" if operation == _INSERT else None
                 _raise_document_too_large(
                     write_op, len(value), max_bson_size)
-            result = send_message()
-            results.append((idx_offset, result))
-            if ordered and "writeErrors" in result:
-                return results
-
-            # Truncate back to the start of list elements
-            buf.seek(list_start + 4)
-            buf.truncate()
-            idx_offset += idx
-            idx = 0
-            key = b'0'
-            to_send = []
+            break
         buf.write(_BSONOBJ)
         buf.write(key)
         buf.write(_ZERO_8)
@@ -799,10 +850,166 @@ def _do_batched_write_command(namespace, operation, command,
         to_send.append(doc)
         idx += 1
 
-    if not has_docs:
-        raise InvalidOperation("cannot do an empty bulk write")
+    # Finalize the current OP_QUERY message.
+    # Close list and command documents
+    buf.write(_ZERO_16)
 
-    results.append((idx_offset, send_message()))
-    return results
+    # Write document lengths and request id
+    length = buf.tell()
+    buf.seek(list_start)
+    buf.write(struct.pack('<i', length - list_start - 1))
+    buf.seek(command_start)
+    buf.write(struct.pack('<i', length - command_start))
+    buf.seek(4)
+    request_id = _randint()
+    buf.write(struct.pack('<i', request_id))
+    buf.seek(0)
+    buf.write(struct.pack('<i', length))
+
+    return request_id, buf.getvalue(), to_send
 if _use_c:
     _do_batched_write_command = _cmessage._do_batched_write_command
+
+
+class _OpReply(object):
+    """A MongoDB OP_REPLY response message."""
+
+    __slots__ = ("flags", "cursor_id", "number_returned", "documents")
+
+    UNPACK = struct.Struct("<iqii").unpack
+    OP_CODE = 1
+
+    def __init__(self, flags, cursor_id, number_returned, documents):
+        self.flags = flags
+        self.cursor_id = cursor_id
+        self.number_returned = number_returned
+        self.documents = documents
+
+    def raw_response(self, cursor_id=None):
+        """Check the response header from the database, without decoding BSON.
+
+        Check the response for errors and unpack.
+
+        Can raise CursorNotFound, NotMasterError, ExecutionTimeout, or
+        OperationFailure.
+
+        :Parameters:
+          - `cursor_id` (optional): cursor_id we sent to get this response -
+            used for raising an informative exception when we get cursor id not
+            valid at server response.
+        """
+        if self.flags & 1:
+            # Shouldn't get this response if we aren't doing a getMore
+            if cursor_id is None:
+                raise ProtocolError("No cursor id for getMore operation")
+
+            # Fake a getMore command response. OP_GET_MORE provides no
+            # document.
+            msg = "Cursor not found, cursor id: %d" % (cursor_id,)
+            errobj = {"ok": 0, "errmsg": msg, "code": 43}
+            raise CursorNotFound(msg, 43, errobj)
+        elif self.flags & 2:
+            error_object = bson.BSON(self.documents).decode()
+            # Fake the ok field if it doesn't exist.
+            error_object.setdefault("ok", 0)
+            if error_object["$err"].startswith("not master"):
+                raise NotMasterError(error_object["$err"], error_object)
+            elif error_object.get("code") == 50:
+                raise ExecutionTimeout(error_object.get("$err"),
+                                       error_object.get("code"),
+                                       error_object)
+            raise OperationFailure("database error: %s" %
+                                   error_object.get("$err"),
+                                   error_object.get("code"),
+                                   error_object)
+        return [self.documents]
+
+    def unpack_response(self, cursor_id=None,
+                        codec_options=_UNICODE_REPLACE_CODEC_OPTIONS):
+        """Unpack a response from the database and decode the BSON document(s).
+
+        Check the response for errors and unpack, returning a dictionary
+        containing the response data.
+
+        Can raise CursorNotFound, NotMasterError, ExecutionTimeout, or
+        OperationFailure.
+
+        :Parameters:
+          - `cursor_id` (optional): cursor_id we sent to get this response -
+            used for raising an informative exception when we get cursor id not
+            valid at server response
+          - `codec_options` (optional): an instance of
+            :class:`~bson.codec_options.CodecOptions`
+        """
+        self.raw_response(cursor_id)
+        return bson.decode_all(self.documents, codec_options)
+
+    def command_response(self):
+        """Unpack a command response."""
+        docs = self.unpack_response()
+        assert self.number_returned == 1
+        return docs[0]
+
+    @classmethod
+    def unpack(cls, msg):
+        """Construct an _OpReply from raw bytes."""
+        # PYTHON-945: ignore starting_from field.
+        flags, cursor_id, _, number_returned = cls.UNPACK(msg[:20])
+        return cls(flags, cursor_id, number_returned, msg[20:])
+
+
+def _first_batch(sock_info, db, coll, query, ntoreturn,
+                 slave_ok, codec_options, read_preference, cmd, listeners):
+    """Simple query helper for retrieving a first (and possibly only) batch."""
+    query = _Query(
+        0, db, coll, 0, query, None, codec_options,
+        read_preference, ntoreturn, 0, DEFAULT_READ_CONCERN, None, None,
+        None)
+
+    name = next(iter(cmd))
+    publish = listeners.enabled_for_commands
+    if publish:
+        start = datetime.datetime.now()
+
+    request_id, msg, max_doc_size = query.get_message(slave_ok, sock_info)
+
+    if publish:
+        encoding_duration = datetime.datetime.now() - start
+        listeners.publish_command_start(
+            cmd, db, request_id, sock_info.address)
+        start = datetime.datetime.now()
+
+    sock_info.send_message(msg, max_doc_size)
+    reply = sock_info.receive_message(request_id)
+    try:
+        docs = reply.unpack_response(None, codec_options)
+    except Exception as exc:
+        if publish:
+            duration = (datetime.datetime.now() - start) + encoding_duration
+            if isinstance(exc, (NotMasterError, OperationFailure)):
+                failure = exc.details
+            else:
+                failure = _convert_exception(exc)
+            listeners.publish_command_failure(
+                duration, failure, name, request_id, sock_info.address)
+        raise
+    # listIndexes
+    if 'cursor' in cmd:
+        result = {
+            u'cursor': {
+                u'firstBatch': docs,
+                u'id': reply.cursor_id,
+                u'ns': u'%s.%s' % (db, coll)
+            },
+            u'ok': 1.0
+        }
+    # fsyncUnlock, currentOp
+    else:
+        result = docs[0] if docs else {}
+        result[u'ok'] = 1.0
+    if publish:
+        duration = (datetime.datetime.now() - start) + encoding_duration
+        listeners.publish_command_success(
+            duration, result, name, request_id, sock_info.address)
+
+    return result
